/* --------------------------------------------- */
/* --------------------------------------------- */
/*  Contains i2c specific hardware information - */
/* --------------------------------------------- */
/* --------------------------------------------- */



struct I2C
	{
	ULONG		Flags	/* see below */
	ULONG		Supported	/* like I2C_Flags */
	ULONG		Size	/* size of this structure (including driver specific extension) */
	STRUCT		Reserved1,8*4
	ULONG		Unit
	APTR		Open	/* function pointer */
	APTR		Close	/* function pointer */
	APTR		Init	/* function pointer */
	APTR		Start	/* function pointer */
	APTR		Stop	/* function pointer */
	APTR		Read	/* function pointer */
	APTR		Write	/* function pointer */
	APTR		Wait	/* function pointer */
	STRUCT		Reserved2,8*4
	LABEL		SIZEOF
	}

/* Flags definitions for I2C_Flags  */

/* bit 0-15 are free for driver specific usage */

#define I2CB_Enable		(16)

#define I2CF_Enable		(1<<I2CB_Enable)



/* ---------------------------------------------- */
/* ---------------------------------------------- */
/*  Contains gpio specific hardware information - */
/* ---------------------------------------------- */
/* ---------------------------------------------- */

struct GPIO
	{
	LONG		Flags	/* see below */
	ULONG		Supported	/* like GPIO_Flags */
	ULONG		Size	/* Size of this structure (including driver specific extension) */
	STRUCT		Reserved1,8*4
	ULONG		Unit
	APTR		Open	/* function pointer */
	APTR		Close	/* function pointer */
	APTR		ReadData	/* function pointer */
	APTR		ModifyData	/* function pointer */
	APTR		ReadOE	/* function pointer */
	APTR		ModifyOE	/* function pointer */
	STRUCT		Reserved2,8*4
	LABEL		SIZEOF
	}

/* Flags definitions for GPIO_Flags  */

/* bit 0-15 are free for driver specific usage */

#define GPIOB_Enable		(16)

#define GPIOF_Enable		(1<<GPIOB_Enable)





/* ---------------------------------------------------------------------------- */
/* ---------------------------------------------------------------------------- */
/*  Contains interrupt specific information.                                  - */
/*                                                                            - */
/*  NOTE: This is an Array. It's required to terminated with a LONG NULL. You - */
/*        can add as many interrupts as required before termination.          - */
/*                                                                            - */
/*        e.g:                                                                - */
/*                                                                            - */
/*        STRUCTURE TVIS,0                                                    - */
/*        STRUCT TVIS1,TVIS_SIZEOF                                            - */
/*        STRUCT TVIS2,TVIS_SIZEOF                                            - */
/*        STRUCT TVIS3,TVIS_SIZEOF                                            - */
/*        ULONG  NULL                                                         - */
/* ---------------------------------------------------------------------------- */
/* ---------------------------------------------------------------------------- */

struct TVIS
	{
	APTR		Name
	APTR		Code
	APTR		Data
	APTR		Reserved1
	APTR		Reserved2
	LABEL		SIZEOF
	}

/* ------------------------------------------------------- */
/* ------------------------------------------------------- */
/*  These values are the return code for the sub drivers - */
/* ------------------------------------------------------- */
/* ------------------------------------------------------- */
#define SubLib_Result_Ready	(0)
#define SubLib_Result_NeedMoreInfo	(1)
#define SubLib_Result_Failed	(2)
#define SubLib_Result_Restart	(3)	/* required if driver changed ID values! */

/* ---------------------------------- */
/* ---------------------------------- */
/*  Some Tag data value Definitions - */
/* ---------------------------------- */
/* ---------------------------------- */
#define HardwareType_Unknown	(0)
#define HardwareType_Virtual	(1)
#define HardwareType_PCI	(2)
#define HardwareType_USB	(3)
#define HardwareType_FireWire	(4)
#define HardwareType_Serial	(5)
#define HardwareType_Parallel	(6)
#define HardwareType_LastType	(6)

/* ------------------------------------------------------------------------ */
/* ------------------------------------------------------------------------ */
/*  Do not touch! Do not touch! Do not touch! Do not touch! Do not touch! - */
/* ------------------------------------------------------------------------ */
/* ------------------------------------------------------------------------ */
/* TAG_USER = $80000000 */
#define NTAG_Empty		(TAG_USER+1)

#define NTAGB_Memory		(30)
#define NTAGB_Protected		(28)
#define NTAGB_Storage		(27)
#define NTAGB_Driver		(26)

#define NTAGF_Memory		(1<<NTAGB_Memory)
#define NTAGF_Protected		(1<<NTAGB_Protected)
#define NTAGF_Storage		(1<<NTAGB_Storage)
#define NTAGF_Driver		(1<<NTAGB_Driver)
/* --- Application Class */
#define NTAG_Data		(TAG_USER+0x1000)
#define NTAG_Memory		(NTAG_Data|NTAGF_Memory)
/* --- Internal Class */
#define NTAGP_Data		(NTAG_Data|NTAGF_Protected)
#define NTAGP_Memory		(NTAG_Memory|NTAGF_Protected)
/* --- Driver Class */
#define NTAGD_Data		(NTAG_Data|NTAGF_Driver)
#define NTAGD_Memory		(NTAG_Memory|NTAGF_Driver)


/* ------------------------------------------------------------------------ */
/* ------------------------------------------------------------------------ */
/*  Internal Tags Internal Tags Internal Tags Internal Tags Internal Tags - */
/* ------------------------------------------------------------------------ */
/* ------------------------------------------------------------------------ */
#define TV_DriverBase		(NTAGP_Data)
#define TV_AudioBase		(NTAGP_Data+1)
#define TV_VideoBase		(NTAGP_Data+2)
#define TV_AVTunerBase		(NTAGP_Data+3)
#define TV_SATTunerBase		(NTAGP_Data+4)
#define TV_RemoteBase		(NTAGP_Data+5)
#define TV_VolumeBase		(NTAGP_Data+6)
#define TV_ReservedBase4	(NTAGP_Data+7)
#define TV_ReservedBase5	(NTAGP_Data+8)
#define TV_ReservedBase6	(NTAGP_Data+9)
#define TV_ReservedBase7	(NTAGP_Data+10)
#define TV_ReservedBase8	(NTAGP_Data+11)

#define TV_DriverAddress	(NTAGP_Data+12)	/* SubLib specific address value (i2c address, memory address or other stuff) */
#define TV_AudioAddress		(NTAGP_Data+13)	/* SubLib specific address value (i2c address, memory address or other stuff) */
#define TV_VideoAddress		(NTAGP_Data+14)	/* SubLib specific address value (i2c address, memory address or other stuff) */
#define TV_AVTunerAddress	(NTAGP_Data+15)	/* SubLib specific address value (i2c address, memory address or other stuff) */
#define TV_SATTunerAddress	(NTAGP_Data+16)	/* SubLib specific address value (i2c address, memory address or other stuff) */
#define TV_RemoteAddress	(NTAGP_Data+17)	/* SubLib specific address value (i2c address, memory address or other stuff) */
#define TV_VolumeAddress	(NTAGP_Data+18)	/* SubLib specific address value (i2c address, memory address or other stuff) */
#define TV_ReservedAddress4	(NTAGP_Data+19)	/* SubLib specific address value (i2c address, memory address or other stuff) */
#define TV_ReservedAddress5	(NTAGP_Data+20)	/* SubLib specific address value (i2c address, memory address or other stuff) */
#define TV_ReservedAddress6	(NTAGP_Data+21)	/* SubLib specific address value (i2c address, memory address or other stuff) */
#define TV_ReservedAddress7	(NTAGP_Data+22)	/* SubLib specific address value (i2c address, memory address or other stuff) */
#define TV_ReservedAddress8	(NTAGP_Data+23)	/* SubLib specific address value (i2c address, memory address or other stuff) */

#define TVI_VendorID		(NTAGP_Data+24)	/* from PCI config space */
#define TVI_DeviceID		(NTAGP_Data+25)	/* from PCI config space */
#define TVI_SubSystemID		(NTAGP_Data+26)	/* from PCI config space */
#define TVI_SubSystemVendorID	(NTAGP_Data+27)	/* from PCI config space */
#define TVI_RevisionID		(NTAGP_Data+28)	/* from PCI config space */
#define TVI_ClassRevision	(NTAGP_Data+29)	/* from PCI config space */
#define TVI_InternalID		(NTAGP_Data+30)
#define TVI_PCIHandle		(NTAGP_Data+31)
#define TVI_BaseMemory		(NTAGP_Data+32)
#define TVI_IRQNumber		(NTAGP_Data+33)
#define TVI_Priority		(NTAGP_Data+34)	/* Application priority */
#define TVI_Number		(NTAGP_Data+35)	/* Card number */
#define TVI_Task		(NTAGP_Data+36)	/* Notification Task */
#define TVI_TaskSignalMask	(NTAGP_Data+37)	/* Notification Signal */
#define TVI_UpdateCard		(NTAGP_Data+38)	/* Taglist to be used during application switch */
#define TV_HardwareType		(NTAGP_Data+39)	/* Virtual, PCI, ... */
#define TV_AutoConfig		(NTAGP_Data+40)	/* Driver modified the data pool driver set. e.g. hauppauge eeprom specifies the currect tuner/audio */
/* ---------------------------------------------------------------------------- */
/* ---------------------------------------------------------------------------- */
#define TV_DriverMemory		(NTAGP_Memory)	/* SubLib specific memory buffer */
#define TV_AudioMemory		(NTAGP_Memory+1)	/* SubLib specific memory buffer */
#define TV_VideoMemory		(NTAGP_Memory+2)	/* SubLib specific memory buffer */
#define TV_AVTunerMemory	(NTAGP_Memory+3)	/* SubLib specific memory buffer */
#define TV_SATTunerMemory	(NTAGP_Memory+4)	/* SubLib specific memory buffer */
#define TV_RemoteMemory		(NTAGP_Memory+5)	/* SubLib specific memory buffer */
#define TV_VolumeMemory		(NTAGP_Memory+6)	/* SubLib specific memory buffer */
#define TV_ReservedMemory2	(NTAGP_Memory+7)	/* SubLib specific memory buffer */
#define TV_ReservedMemory3	(NTAGP_Memory+8)	/* SubLib specific memory buffer */
#define TV_ReservedMemory4	(NTAGP_Memory+9)	/* SubLib specific memory buffer */
#define TV_ReservedMemory5	(NTAGP_Memory+10)	/* SubLib specific memory buffer */
#define TV_ReservedMemory6	(NTAGP_Memory+11)	/* SubLib specific memory buffer */
#define TV_ReleaseCode		(NTAGP_Memory+12)	/* PCI Release Code */

#define TV_DriverName		(NTAGP_Memory+13)	/* SubLib Name */
#define TV_AudioName		(NTAGP_Memory+14)	/* SubLib Name */
#define TV_VideoName		(NTAGP_Memory+15)	/* SubLib Name */
#define TV_AVTunerName		(NTAGP_Memory+16)	/* SubLib Name */
#define TV_SATTunerName		(NTAGP_Memory+17)	/* SubLib Name */
#define TV_RemoteName		(NTAGP_Memory+18)	/* SubLib Name */
#define TV_VolumeName		(NTAGP_Memory+19)	/* SubLib Name */
#define TV_ReservedName2	(NTAGP_Memory+20)	/* SubLib Name */
#define TV_ReservedName3	(NTAGP_Memory+21)	/* SubLib Name */
#define TV_ReservedName4	(NTAGP_Memory+22)	/* SubLib Name */
#define TV_ReservedName5	(NTAGP_Memory+23)	/* SubLib Name */
#define TV_ReservedName6	(NTAGP_Memory+24)	/* SubLib Name */
#define TVI_CardName		(NTAGP_Memory+25)	/* PCICard Name */
#define TVI_VendorName		(NTAGP_Memory+26)	/* Vendor Name */
#define TVI_SubSystemVendorName	(NTAGP_Memory+27)	/* SubSystemVendor Name */
#define TV_HardwareData		(NTAGP_Memory+28)	/* Hardware Data (e.g. unit number, ...) */
/* ---------------------------------------------------------------------------- */
/*  DRIVER TAGS DRIVER TAGS DRIVER TAGS DRIVER TAGS DRIVER TAGS DRIVER TAGS DRI */
/* ---------------------------------------------------------------------------- */
/* ################## INSERT HERE #################### */
#define TV_VideoGPIOInitData	(NTAGD_Data)
#define TV_VideoGPIOInitOE	(NTAGD_Data+1)
/* ---------------------------------------------------------------------------- */
/* ################## INSERT HERE #################### */
#define TV_I2C			(NTAGD_Memory)
#define TV_DriverConfig		(NTAGD_Memory+1)	/* SubLib specific config space (see CardData) */
#define TV_AudioConfig		(NTAGD_Memory+2)	/* SubLib specific config space (see CardData) */
#define TV_VideoConfig		(NTAGD_Memory+3)	/* SubLib specific config space (see CardData) */
#define TV_AVTunerConfig	(NTAGD_Memory+4)	/* SubLib specific config space (see CardData) */
#define TV_SATTunerConfig	(NTAGD_Memory+5)	/* SubLib specific config space (see CardData) */
#define TV_RemoteConfig		(NTAGD_Memory+6)	/* SubLib specific config space (see CardData) */
#define TV_VolumeConfig		(NTAGD_Memory+7)	/* SubLib specific config space (see CardData) */
#define TV_ReservedConfig2	(NTAGD_Memory+8)	/* SubLib specific config space (see CardData) */
#define TV_ReservedConfig3	(NTAGD_Memory+9)	/* SubLib specific config space (see CardData) */
#define TV_ReservedConfig4	(NTAGD_Memory+10)	/* SubLib specific config space (see CardData) */
#define TV_ReservedConfig5	(NTAGD_Memory+11)	/* SubLib specific config space (see CardData) */
#define TV_ReservedConfig6	(NTAGD_Memory+12)	/* SubLib specific config space (see CardData) */

#define TV_DriverInterrupts	(NTAGD_Memory+13)	/* SubLib interrupt structure */
#define TV_AudioInterrupts	(NTAGD_Memory+14)	/* SubLib interrupt structure */
#define TV_VideoInterrupts	(NTAGD_Memory+15)	/* SubLib interrupt structure */
#define TV_AVTunerInterrupts	(NTAGD_Memory+16)	/* SubLib interrupt structure */
#define TV_SATTunerInterrupts	(NTAGD_Memory+17)	/* SubLib interrupt structure */
#define TV_RemoteInterrupts	(NTAGD_Memory+18)	/* SubLib interrupt structure */
#define TV_VolumeInterrupts	(NTAGD_Memory+19)	/* SubLib interrupt structure */
#define TV_ReservedInterrupts2	(NTAGD_Memory+20)	/* SubLib interrupt structure */
#define TV_ReservedInterrupts3	(NTAGD_Memory+21)	/* SubLib interrupt structure */
#define TV_ReservedInterrupts4	(NTAGD_Memory+22)	/* SubLib interrupt structure */
#define TV_ReservedInterrupts5	(NTAGD_Memory+23)	/* SubLib interrupt structure */
#define TV_ReservedInterrupts6	(NTAGD_Memory+24)	/* SubLib interrupt structure */

#define TV_EEpromMemory		(NTAGD_Memory+25)
#define TV_GPIO			(NTAGD_Memory+26)
#define TV_PCI			(NTAGD_Memory+27)
