;---------------------------------------------
;---------------------------------------------
; Contains i2c specific hardware information -
;---------------------------------------------
;---------------------------------------------



      STRUCTURE I2C,0
      ULONG     I2C_Flags         ;see below
      ULONG     I2C_Supported     ;like I2C_Flags
      ULONG     I2C_Size          ;size of this structure (including driver specific extension)
      STRUCT    I2C_Reserved1,8*4
      ULONG     I2C_Unit
      APTR      I2C_Open          ;function pointer
      APTR      I2C_Close         ;function pointer
      APTR      I2C_Init          ;function pointer
      APTR      I2C_Start         ;function pointer
      APTR      I2C_Stop          ;function pointer
      APTR      I2C_Read          ;function pointer
      APTR      I2C_Write         ;function pointer
      APTR      I2C_Wait          ;function pointer
      STRUCT    I2C_Reserved2,8*4
      LABEL     I2C_SIZEOF

;Flags definitions for I2C_Flags 

;bit 0-15 are free for driver specific usage

I2CB_Enable             = 16

I2CF_Enable             = 1<<I2CB_Enable



;----------------------------------------------
;----------------------------------------------
; Contains gpio specific hardware information -
;----------------------------------------------
;----------------------------------------------

      STRUCTURE GPIO,0
      LONG      GPIO_Flags       ;see below
      ULONG     GPIO_Supported   ;like GPIO_Flags
      ULONG     GPIO_Size        ;Size of this structure (including driver specific extension)
      STRUCT    GPIO_Reserved1,8*4
      ULONG     GPIO_Unit
      APTR      GPIO_Open        ;function pointer
      APTR      GPIO_Close       ;function pointer
      APTR      GPIO_ReadData    ;function pointer
      APTR      GPIO_ModifyData  ;function pointer
      APTR      GPIO_ReadOE      ;function pointer
      APTR      GPIO_ModifyOE    ;function pointer
      STRUCT    GPIO_Reserved2,8*4
      LABEL     GPIO_SIZEOF

;Flags definitions for GPIO_Flags 

;bit 0-15 are free for driver specific usage

GPIOB_Enable             = 16

GPIOF_Enable             = 1<<GPIOB_Enable





;----------------------------------------------------------------------------
;----------------------------------------------------------------------------
; Contains interrupt specific information.                                  -
;                                                                           -
; NOTE: This is an Array. It's required to terminated with a LONG NULL. You -
;       can add as many interrupts as required before termination.          -
;                                                                           -
;       e.g:                                                                -
;                                                                           -
;       STRUCTURE TVIS,0                                                    -
;       STRUCT TVIS1,TVIS_SIZEOF                                            -
;       STRUCT TVIS2,TVIS_SIZEOF                                            -
;       STRUCT TVIS3,TVIS_SIZEOF                                            -
;       ULONG  NULL                                                         -
;----------------------------------------------------------------------------
;----------------------------------------------------------------------------

      STRUCTURE TVIS,0
      APTR      TVIS_Name
      APTR      TVIS_Code
      APTR      TVIS_Data
      APTR      TVIS_Reserved1
      APTR      TVIS_Reserved2
      LABEL     TVIS_SIZEOF

;-------------------------------------------------------
;-------------------------------------------------------
; These values are the return code for the sub drivers -
;-------------------------------------------------------
;-------------------------------------------------------
SubLib_Result_Ready             = 0
SubLib_Result_NeedMoreInfo      = 1
SubLib_Result_Failed            = 2
SubLib_Result_Restart           = 3 ;required if driver changed ID values!

;----------------------------------
;----------------------------------
; Some Tag data value Definitions -
;----------------------------------
;----------------------------------
HardwareType_Unknown    = 0
HardwareType_Virtual    = 1
HardwareType_PCI        = 2
HardwareType_USB        = 3
HardwareType_FireWire   = 4
HardwareType_Serial     = 5
HardwareType_Parallel   = 6
HardwareType_LastType   = 6

;------------------------------------------------------------------------
;------------------------------------------------------------------------
; Do not touch! Do not touch! Do not touch! Do not touch! Do not touch! -
;------------------------------------------------------------------------
;------------------------------------------------------------------------
;TAG_USER = $80000000
NTAG_Empty        = TAG_USER+1

NTAGB_Memory      = 30
NTAGB_Protected   = 28
NTAGB_Storage     = 27
NTAGB_Driver      = 26

NTAGF_Memory      = 1<<NTAGB_Memory
NTAGF_Protected   = 1<<NTAGB_Protected
NTAGF_Storage     = 1<<NTAGB_Storage
NTAGF_Driver      = 1<<NTAGB_Driver
;--- Application Class
NTAG_Data         = TAG_USER+$1000
NTAG_Memory       = NTAG_Data|NTAGF_Memory
;--- Internal Class
NTAGP_Data        = NTAG_Data|NTAGF_Protected
NTAGP_Memory      = NTAG_Memory|NTAGF_Protected
;--- Driver Class
NTAGD_Data        = NTAG_Data|NTAGF_Driver
NTAGD_Memory      = NTAG_Memory|NTAGF_Driver


;------------------------------------------------------------------------
;------------------------------------------------------------------------
; Internal Tags Internal Tags Internal Tags Internal Tags Internal Tags -
;------------------------------------------------------------------------
;------------------------------------------------------------------------
TV_DriverBase            = NTAGP_Data
TV_AudioBase             = NTAGP_Data+1
TV_VideoBase             = NTAGP_Data+2
TV_AVTunerBase           = NTAGP_Data+3
TV_SATTunerBase          = NTAGP_Data+4
TV_RemoteBase            = NTAGP_Data+5
TV_VolumeBase            = NTAGP_Data+6
TV_ReservedBase4         = NTAGP_Data+7
TV_ReservedBase5         = NTAGP_Data+8
TV_ReservedBase6         = NTAGP_Data+9
TV_ReservedBase7         = NTAGP_Data+10
TV_ReservedBase8         = NTAGP_Data+11

TV_DriverAddress         = NTAGP_Data+12   ;SubLib specific address value (i2c address, memory address or other stuff)
TV_AudioAddress          = NTAGP_Data+13   ;SubLib specific address value (i2c address, memory address or other stuff)
TV_VideoAddress          = NTAGP_Data+14   ;SubLib specific address value (i2c address, memory address or other stuff)
TV_AVTunerAddress        = NTAGP_Data+15   ;SubLib specific address value (i2c address, memory address or other stuff)
TV_SATTunerAddress       = NTAGP_Data+16   ;SubLib specific address value (i2c address, memory address or other stuff)
TV_RemoteAddress         = NTAGP_Data+17   ;SubLib specific address value (i2c address, memory address or other stuff)
TV_VolumeAddress         = NTAGP_Data+18   ;SubLib specific address value (i2c address, memory address or other stuff)
TV_ReservedAddress4      = NTAGP_Data+19   ;SubLib specific address value (i2c address, memory address or other stuff)
TV_ReservedAddress5      = NTAGP_Data+20   ;SubLib specific address value (i2c address, memory address or other stuff)
TV_ReservedAddress6      = NTAGP_Data+21   ;SubLib specific address value (i2c address, memory address or other stuff)
TV_ReservedAddress7      = NTAGP_Data+22   ;SubLib specific address value (i2c address, memory address or other stuff)
TV_ReservedAddress8      = NTAGP_Data+23   ;SubLib specific address value (i2c address, memory address or other stuff)

TVI_VendorID             = NTAGP_Data+24   ;from PCI config space
TVI_DeviceID             = NTAGP_Data+25   ;from PCI config space
TVI_SubSystemID          = NTAGP_Data+26   ;from PCI config space
TVI_SubSystemVendorID    = NTAGP_Data+27   ;from PCI config space
TVI_RevisionID           = NTAGP_Data+28   ;from PCI config space
TVI_ClassRevision        = NTAGP_Data+29   ;from PCI config space
TVI_InternalID           = NTAGP_Data+30
TVI_PCIHandle            = NTAGP_Data+31
TVI_BaseMemory           = NTAGP_Data+32
TVI_IRQNumber            = NTAGP_Data+33
TVI_Priority             = NTAGP_Data+34   ;Application priority
TVI_Number               = NTAGP_Data+35   ;Card number
TVI_Task                 = NTAGP_Data+36   ;Notification Task
TVI_TaskSignalMask       = NTAGP_Data+37   ;Notification Signal
TVI_UpdateCard           = NTAGP_Data+38   ;Taglist to be used during application switch
TV_HardwareType          = NTAGP_Data+39   ;Virtual, PCI, ...
TV_AutoConfig            = NTAGP_Data+40   ;Driver modified the data pool driver set. e.g. hauppauge eeprom specifies the currect tuner/audio
;----------------------------------------------------------------------------
;----------------------------------------------------------------------------
TV_DriverMemory          = NTAGP_Memory    ;SubLib specific memory buffer
TV_AudioMemory           = NTAGP_Memory+1  ;SubLib specific memory buffer
TV_VideoMemory           = NTAGP_Memory+2  ;SubLib specific memory buffer
TV_AVTunerMemory         = NTAGP_Memory+3  ;SubLib specific memory buffer
TV_SATTunerMemory        = NTAGP_Memory+4  ;SubLib specific memory buffer
TV_RemoteMemory          = NTAGP_Memory+5  ;SubLib specific memory buffer
TV_VolumeMemory          = NTAGP_Memory+6  ;SubLib specific memory buffer
TV_ReservedMemory2       = NTAGP_Memory+7  ;SubLib specific memory buffer
TV_ReservedMemory3       = NTAGP_Memory+8  ;SubLib specific memory buffer
TV_ReservedMemory4       = NTAGP_Memory+9  ;SubLib specific memory buffer
TV_ReservedMemory5       = NTAGP_Memory+10 ;SubLib specific memory buffer
TV_ReservedMemory6       = NTAGP_Memory+11 ;SubLib specific memory buffer
TV_ReleaseCode           = NTAGP_Memory+12 ;PCI Release Code

TV_DriverName            = NTAGP_Memory+13 ;SubLib Name
TV_AudioName             = NTAGP_Memory+14 ;SubLib Name
TV_VideoName             = NTAGP_Memory+15 ;SubLib Name
TV_AVTunerName           = NTAGP_Memory+16 ;SubLib Name
TV_SATTunerName          = NTAGP_Memory+17 ;SubLib Name
TV_RemoteName            = NTAGP_Memory+18 ;SubLib Name
TV_VolumeName            = NTAGP_Memory+19 ;SubLib Name
TV_ReservedName2         = NTAGP_Memory+20 ;SubLib Name
TV_ReservedName3         = NTAGP_Memory+21 ;SubLib Name
TV_ReservedName4         = NTAGP_Memory+22 ;SubLib Name
TV_ReservedName5         = NTAGP_Memory+23 ;SubLib Name
TV_ReservedName6         = NTAGP_Memory+24 ;SubLib Name
TVI_CardName             = NTAGP_Memory+25 ;PCICard Name
TVI_VendorName           = NTAGP_Memory+26 ;Vendor Name
TVI_SubSystemVendorName  = NTAGP_Memory+27 ;SubSystemVendor Name
TV_HardwareData          = NTAGP_Memory+28 ;Hardware Data (e.g. unit number, ...)
;----------------------------------------------------------------------------
; DRIVER TAGS DRIVER TAGS DRIVER TAGS DRIVER TAGS DRIVER TAGS DRIVER TAGS DRI
;----------------------------------------------------------------------------
;################## INSERT HERE ####################
TV_VideoGPIOInitData     = NTAGD_Data
TV_VideoGPIOInitOE       = NTAGD_Data+1
;----------------------------------------------------------------------------
;################## INSERT HERE ####################
TV_I2C                   = NTAGD_Memory
TV_DriverConfig          = NTAGD_Memory+1  ;SubLib specific config space (see CardData)
TV_AudioConfig           = NTAGD_Memory+2  ;SubLib specific config space (see CardData)
TV_VideoConfig           = NTAGD_Memory+3  ;SubLib specific config space (see CardData)
TV_AVTunerConfig         = NTAGD_Memory+4  ;SubLib specific config space (see CardData)
TV_SATTunerConfig        = NTAGD_Memory+5  ;SubLib specific config space (see CardData)
TV_RemoteConfig          = NTAGD_Memory+6  ;SubLib specific config space (see CardData)
TV_VolumeConfig          = NTAGD_Memory+7  ;SubLib specific config space (see CardData)
TV_ReservedConfig2       = NTAGD_Memory+8  ;SubLib specific config space (see CardData)
TV_ReservedConfig3       = NTAGD_Memory+9  ;SubLib specific config space (see CardData)
TV_ReservedConfig4       = NTAGD_Memory+10 ;SubLib specific config space (see CardData)
TV_ReservedConfig5       = NTAGD_Memory+11 ;SubLib specific config space (see CardData)
TV_ReservedConfig6       = NTAGD_Memory+12 ;SubLib specific config space (see CardData)

TV_DriverInterrupts      = NTAGD_Memory+13 ;SubLib interrupt structure
TV_AudioInterrupts       = NTAGD_Memory+14 ;SubLib interrupt structure
TV_VideoInterrupts       = NTAGD_Memory+15 ;SubLib interrupt structure
TV_AVTunerInterrupts     = NTAGD_Memory+16 ;SubLib interrupt structure
TV_SATTunerInterrupts    = NTAGD_Memory+17 ;SubLib interrupt structure
TV_RemoteInterrupts      = NTAGD_Memory+18 ;SubLib interrupt structure
TV_VolumeInterrupts      = NTAGD_Memory+19 ;SubLib interrupt structure
TV_ReservedInterrupts2   = NTAGD_Memory+20 ;SubLib interrupt structure
TV_ReservedInterrupts3   = NTAGD_Memory+21 ;SubLib interrupt structure
TV_ReservedInterrupts4   = NTAGD_Memory+22 ;SubLib interrupt structure
TV_ReservedInterrupts5   = NTAGD_Memory+23 ;SubLib interrupt structure
TV_ReservedInterrupts6   = NTAGD_Memory+24 ;SubLib interrupt structure

TV_EEpromMemory          = NTAGD_Memory+25
TV_GPIO                  = NTAGD_Memory+26
TV_PCI                   = NTAGD_Memory+27
